@@ -1,5 +1,5 @@
 from logging import getLogger
-from typing import AnyStr, Dict, List, Optional, Union
+from typing import AnyStr, Callable, Dict, List, Optional, Tuple, Union
 
 from .api_resources.mutations import *
 from .api_resources.payload import (
@@ -9,12 +9,20 @@ from .api_resources.payload import (
 )
 from .api_resources.requests import GraphqlRequest, PlainRequest, RestRequest
 from .http import HTTPClient
-from .utils import remove_empty_items
 
 
 logger = getLogger(__name__)
 
 
+def _get_return(res: Dict, get_data: Callable[[Dict], Union[Dict, List]]):
+    """Process response and get data from path if provided."""
+    _res = res
+    if "response" in res and "headers" in res:
+        _res = res["response"]
+        return {"response": get_data(_res), "headers": res["headers"]}
+    return get_data(_res)
+
+
 class Anvil:
     """Main Anvil API class.
 
@@ -26,23 +34,25 @@ class Anvil:
         >> pdf_data = anvil.fill_pdf("the_template_id", payload)
     """
 
-    def __init__(self, api_key=None):
-        self.client = HTTPClient(api_key=api_key)
+    def __init__(self, api_key=None, environment='dev'):
+        self.client = HTTPClient(api_key=api_key, environment=environment)
 
-    def query(self, query: str, variables: Optional[str] = None):
+    def query(self, query: str, variables: Optional[str] = None, **kwargs):
         gql = GraphqlRequest(client=self.client)
-        return gql.post(query, variables=variables)
+        return gql.post(query, variables=variables, **kwargs)
 
-    def mutate(self, query: BaseQuery, variables: dict):
+    def mutate(self, query: BaseQuery, variables: dict, **kwargs):
         gql = GraphqlRequest(client=self.client)
-        return gql.post(query.get_mutation(), variables)
+        return gql.post(query.get_mutation(), variables, **kwargs)
 
     def request_rest(self, options: Optional[dict] = None):
         api = RestRequest(self.client, options=options)
         return api
 
-    def fill_pdf(self, template_id: str, payload: Union[dict, AnyStr]):
-        """Fills an existing template with provided payload data.
+    def fill_pdf(
+        self, template_id: str, payload: Union[dict, AnyStr, FillPDFPayload], **kwargs
+    ):
+        """Fill an existing template with provided payload data.
 
         Use the casts graphql query to get a list of available templates you
         can use for this request.
@@ -52,100 +62,98 @@ class Anvil:
         :param payload: payload in the form of a dict or JSON data
         :type payload: dict|str
         """
-        data = None
         try:
             if isinstance(payload, dict):
-                data = FillPDFPayload.from_dict(payload)
+                data = FillPDFPayload(**payload)
             elif isinstance(payload, str):
-                data = FillPDFPayload.from_json(payload)
+                data = FillPDFPayload.parse_raw(
+                    payload, content_type="application/json"
+                )
             elif isinstance(payload, FillPDFPayload):
                 data = payload
             else:
                 raise ValueError("`payload` must be a valid JSON string or a dict")
         except KeyError as e:
+            logger.exception(e)
             raise ValueError(
                 "`payload` validation failed. Please make sure all required "
                 "fields are set. "
-            )
+            ) from e
 
         api = RestRequest(client=self.client)
         return api.post(
             f"fill/{template_id}.pdf",
-            remove_empty_items(data.to_dict() if data else {}),
+            data.dict(by_alias=True, exclude_none=True) if data else {},
+            **kwargs,
         )
 
-    def generate_pdf(self, payload: Union[AnyStr, Dict]):
+    def generate_pdf(self, payload: Union[AnyStr, Dict], **kwargs):
         if not payload:
             raise ValueError("`payload` must be a valid JSON string or a dict")
 
         if isinstance(payload, dict):
-            data = GeneratePDFPayload.from_dict(payload)
+            data = GeneratePDFPayload(**payload)
         elif isinstance(payload, str):
-            data = GeneratePDFPayload.from_json(payload)
+            data = GeneratePDFPayload.parse_raw(
+                payload, content_type="application/json"
+            )
         else:
             raise ValueError("`payload` must be a valid JSON string or a dict")
 
-        # `dataclasses_json` doesn't seem to validate the `type` field here,
-        # so double-check that.
-        if data.type and data.type.lower() not in ["html", "markdown"]:
-            raise ValueError("`type` must be either 'html' or 'markdown'")
-
-        # For HTML-type payloads, data must be a dict (or JSON object).
-        if data.type == "html":
-            if not isinstance(data.data, dict):
-                raise ValueError(
-                    "`payload` data must be a dict when generating a PDF from HTML"
-                )
-            if "html" not in data.data:
-                raise ValueError(
-                    "`payload` data must have 'html' if using the 'html' type"
-                )
-        elif data.type == 'markdown' and not isinstance(data.data, list):
-            raise ValueError(
-                "`payload` data must be a list of dicts when generating a PDF "
-                "from markdown (this is the default)"
-            )
-
         # Any data errors would come from here..
         api = RestRequest(client=self.client)
-        return api.post("generate-pdf", data=remove_empty_items(data.to_dict()))
+        return api.post(
+            "generate-pdf", data=data.dict(by_alias=True, exclude_none=True), **kwargs
+        )
 
-    def get_cast(self, eid: str, fields=None):
+    def get_cast(self, eid: str, fields=None, **kwargs):
         if not fields:
             # Use default fields
             fields = ['eid', 'title', 'fieldInfo']
 
         res = self.query(
             f"""{{
-          cast(eid: "{eid}") {{
-            {" ".join(fields)}
-          }}
-        }}"""
+              cast(eid: "{eid}") {{
+                {" ".join(fields)}
+              }}
+            }}""",
+            **kwargs,
         )
 
-        return res["data"]["cast"]
+        def get_data(r):
+            return r["data"]["cast"]
 
-    def get_casts(self, fields=None) -> List:
+        return _get_return(res, get_data=get_data)
+
+    def get_casts(
+        self, fields=None, show_all=False, **kwargs
+    ) -> Union[List, Tuple[List, Dict]]:
         if not fields:
             # Use default fields
-            fields = ['eid', 'title', 'fieldInfo']
+            fields = ["eid", "title", "fieldInfo"]
+
+        cast_args = "" if show_all else "(isTemplate: true)"
 
         res = self.query(
             f"""{{
               currentUser {{
                 organizations {{
-                  casts {{
+                  casts {cast_args} {{
                     {" ".join(fields)}
                   }}
                 }}
               }}
-            }}"""
+            }}""",
+            **kwargs,
         )
 
-        orgs = res["data"]["currentUser"]["organizations"]
-        return [item for org in orgs for item in org["casts"]]
+        def get_data(r):
+            orgs = r["data"]["currentUser"]["organizations"]
+            return [item for org in orgs for item in org["casts"]]
 
-    def get_current_user(self):
+        return _get_return(res, get_data=get_data)
+
+    def get_current_user(self, **kwargs):
         res = self.query(
             """{
               currentUser {
@@ -164,12 +172,13 @@ class Anvil:
                 }
               }
             }
-            """
+            """,
+            **kwargs,
         )
-        user = res["data"]["currentUser"]
-        return user
 
-    def get_welds(self) -> list:
+        return _get_return(res, get_data=lambda r: r["data"]["currentUser"])
+
+    def get_welds(self, **kwargs) -> Union[List, Tuple[List, Dict]]:
         res = self.query(
             """{
               currentUser {
@@ -181,40 +190,28 @@ class Anvil:
                   }
                 }
               }
-            }"""
-        )
-        orgs = res["data"]["currentUser"]["organizations"]
-        return [item for org in orgs for item in org["welds"]]
-
-    def get_queries(self):
-        """
-        Gets list of available queries.
-
-        TODO: This is probably going to be removed in the near future
-            (in the API).
-        :return:
-        """
-        res = self.query(
-            """{
-              __schema {
-                queryType {
-                  fields {
-                    name
-                    description
-                  }
-                }
-              }
-            }"""
+            }""",
+            **kwargs,
         )
 
-        return res["data"]["__schema"]["queryType"]
+        def get_data(r):
+            orgs = r["data"]["currentUser"]["organizations"]
+            return [item for org in orgs for item in org["welds"]]
+
+        return _get_return(res, get_data=get_data)
 
     def create_etch_packet(
         self,
         payload: Optional[
-            Union[dict, CreateEtchPacketPayload, CreateEtchPacket]
+            Union[
+                dict,
+                CreateEtchPacketPayload,
+                CreateEtchPacket,
+                AnyStr,
+            ]
         ] = None,
         json=None,
+        **kwargs,
     ):
         """Create etch packet via a graphql mutation."""
         # Create an etch packet payload instance excluding signers and files
@@ -223,7 +220,9 @@ class Anvil:
             raise TypeError('One of the arguments `payload` or `json` must exist')
 
         if json:
-            payload = CreateEtchPacketPayload.from_json(json)
+            payload = CreateEtchPacketPayload.parse_raw(
+                json, content_type="application/json"
+            )
 
         if isinstance(payload, dict):
             mutation = CreateEtchPacket.create_from_dict(payload)
@@ -235,19 +234,22 @@ class Anvil:
             raise ValueError(
                 "`payload` must be a valid CreateEtchPacket instance or dict"
             )
+        return self.mutate(
+            mutation,
+            variables=mutation.create_payload().dict(by_alias=True, exclude_none=True),
+            **kwargs,
+        )
 
-        return self.mutate(mutation, variables=mutation.create_payload().to_dict())
-
-    def generate_etch_signing_url(self, signer_eid: str, client_user_id: str):
+    def generate_etch_signing_url(self, signer_eid: str, client_user_id: str, **kwargs):
         """Generate a signing URL for a given user."""
         mutation = GenerateEtchSigningURL(
             signer_eid=signer_eid,
             client_user_id=client_user_id,
         )
         payload = mutation.create_payload()
-        return self.mutate(mutation, variables=payload.to_dict())
+        return self.mutate(mutation, variables=payload.dict(by_alias=True), **kwargs)
 
-    def download_documents(self, document_group_eid: str):
+    def download_documents(self, document_group_eid: str, **kwargs):
         """Retrieve all completed documents in zip form."""
         api = PlainRequest(client=self.client)
-        return api.get(f"document-group/{document_group_eid}.zip")
+        return api.get(f"document-group/{document_group_eid}.zip", **kwargs)
